/*
 * bsp.c
 *
 * BSP compiler.
 *
 * Format of input line list file:
 *
 *      %d,%d-%d,%d             x1,y1-x2,y2
 *
 * The list is terminated when data is: 0,0-0,0.
 */

#include <dpmi.h>
#include <go32.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/farptr.h>
#include "bsp.h"


/*****************************************************************************
 Data
*****************************************************************************/

/*
 * Line List
 */

static LINE *line_list = NULL;
static int  line_id = 0;

/*
 * BSP Tree
 */

static BSP  *tree;


/*****************************************************************************
 BSP Construction and Manipulation
*****************************************************************************/

/*
 * AddLineToEnd(): Adds a line to the end of the original line list
 */

static void AddLineToEnd(LINE *to, LINE *l)
{
    if (to->next == NULL)
    {
        to->next = l;
        return;
    }

    AddLineToEnd(to->next, l);
}

/*
 * KillList(): Recursively deletes a linked list of lines
 */

static void KillList(LINE *l)
{
    if (l == NULL)
        return;

    KillList(l->next);

    free(l);
}

/*
 * AddLineToList(): Adds a new line to the linked list. For simplicity, the
 * line is simply inserted at the front. Returns non-zero if out of memory
 */

static int AddLineToList(LINE **linelist, LINE *line)
{
    LINE    *new;

    if ((new = malloc(sizeof(LINE))) == NULL)
        return 1;

    *new = *line;

    if (*linelist == NULL)  // this will be the first element
        new->next = NULL;
    else
        new->next = *linelist;
    *linelist = new;

    return 0;
}

/*
 * BSP Tree Construction:
 * ----------------------
 *
 * Sample world:
 *
 *      ------------------------------- B
 *      ------------------------------- C
 *      ---------------------------     A (root)
 *          ----------------            D
 *
 * A is the root. Assume the walls are defined in a linear list in
 * alphabetical order. This is important because notice that C is closer to
 * A than B, even though B appears next in the list.
 *
 * A good tree for this might look like (front on left, back on right):
 *
 *                  A
 *                 / \
 *                C   D
 *               /
 *              B
 *
 * An equivalent tree (which our BSP construction algorithm ends up building)
 * is:
 *
 *                  A
 *                 / \
 *                B   D
 *                 \
 *                  C
 *
 * This new tree is traversed like this: There is no front leaf for B, so
 * nothing is drawn. B is drawn. Then, its back leaf, C, is drawn. Then, A
 * is drawn, and finally, A's back leaf, D, is drawn. So, B->C->A->D. The same
 * as the first tree!
 *
 * Trees are built using this recursive process:
 *
 *      - A root is selected out of the input list of lines
 *      - Front and back lists are created. These are simply linear lists
 *          - Walls in front of the root's splitting lines are put in the
 *            front list as they are encountered, walls in the back are put
 *            in the back list
 *          - If a wall is split by the splitting line, the front part is
 *            put in the front list, the back part goes in the back list
 *      - A root BSP node is created. The front leaf is the front list (thus
 *        the first element in the front list) and the back leaf is the back
 *        list.
 *      - Recursively repeat the process using the front leaf as a root
 *      - Recursively repeat the process using the back leaf as a root
 *
 * NOTE: Each time BuildBSPTree() is called, a list of usable lines is passed.
 * When it is called recursively for, say, the front list, only those lines
 * can be valid for use. This ensures that space is subdivided correctly. The
 * key is to only scan the lines in the parent's subspace, not the whole list.
 */

/*
 * BuildBSPTree(): Builds a BSP tree recursively. tree must be a pointer to a
 * NULL pointer and rootline must be valid. Returns non-zero if memory runs
 * out
 */

static int BuildBSPTree(BSP **tree, LINE *rootline, LINE *lines)
{
    LINE        *frontlist, *backlist, *current;
    double      norm_x, norm_y, numer, denom, tintersect;
    int         error_code;

    /*
     * Calculate normal to the splitter (which is the infinite root line) by
     * exchanging the X and Y lengths of the root line segment and negating
     * the Y
     */

    norm_x = (rootline->y1 - rootline->y2);
    norm_y = -(rootline->x1 - rootline->x2);

    /*
     * Categorize all lines (except the root) as either in front or behind the
     * the root and add them to the appropriate list
     */

    frontlist = backlist = NULL;

    current = lines;
    while (current != NULL)
    {
        if (current == rootline)    // skip the root line
        {
            current = current->next;
            continue;
        }

        /*
         * Calculate the numerator and denominator of the parametric
         * intersection formula
         */

        numer = norm_x * (current->x1 - rootline->x1) + norm_y * (current->y1 - rootline->y1);
        denom = (-norm_x) * (current->x2 - current->x1) + (-norm_y) * (current->y2 - current->y1);

        /*
         * Now see if the splitter and the current line segment intersect
         */

        if (denom == 0) // if denominator is 0, lines are parallel
        {
            LINE    **l;

            if (numer < 0)  // current line is in front of root
                l = &frontlist;
            else            // current line is behind root
                l = &backlist;

            if (AddLineToList(l, current))
            {
                KillList(frontlist);
                KillList(backlist);
                return 1;   // error (BSP tree must be killed by caller)
            }
        }
        else            // there may or may not be an intersection
        {
            LINE    **l;

            tintersect = numer / denom;

            if ((tintersect <= 0.0) || (tintersect >= 1.0))
            {
                if (numer < -TOLERANCE)     // current line is in front of root
                    l = &frontlist;
                else if (numer > TOLERANCE) // current line is behind root
                   l = &backlist;
                else
                {
                    numer = norm_x * (current->x2 - rootline->x1) + norm_y * (current->y2 - rootline->y1);
                    if (numer < 0)
                        l = &frontlist;
                    else
                        l = &backlist;
                }

                if (AddLineToList(l, current))
                {
                    KillList(frontlist);
                    KillList(backlist);
                    return 1;   // error (BSP tree must be killed by caller)
                }
            }
            else                                    // intersection!
            {
                LINE    *splitline;

                if ((splitline = malloc(sizeof(LINE))) == NULL)
                {
                    KillList(frontlist);
                    KillList(backlist);
                    return 1;
                }

                splitline->x1 = (double) current->x1 + (double) (current->x2 - current->x1) * tintersect;
                splitline->y1 = (double) current->y1 + (double) (current->y2 - current->y1) * tintersect;
                splitline->x2 = current->x2;
                splitline->y2 = current->y2;
                splitline->id = line_id++;
                splitline->next = current->next;    // insert right after currentline

                current->next = splitline;
                current->x2 = (double) current->x2 - (double) (current->x2 - current->x1) * (1.0 - tintersect);
                current->y2 = (double) current->y2 - (double) (current->y2 - current->y1) * (1.0 - tintersect);

                if (numer < 0)  // split part is behind root
                {
                    error_code = AddLineToList(&backlist, splitline);
                    error_code |= AddLineToList(&frontlist, current);
                }
                else            // split part is in front of root
                {
                    error_code = AddLineToList(&frontlist, splitline);
                    error_code |= AddLineToList(&backlist, current);
                }

                if (error_code)
                {
                    KillList(frontlist);
                    KillList(backlist);
                    return 1;   // error (BSP tree must be killed by caller)
                }

                current = current->next;    // advance twice
            }
        }

        current = current->next;
    }

    /*
     * Create a BSP tree node out of the root line and add the front and back
     * lists to it
     */

    error_code = 0;
    if ((*tree = malloc(sizeof(BSP))) != NULL)
    {
        (*tree)->id = rootline->id;

        (*tree)->x1 = rootline->x1;
        (*tree)->y1 = rootline->y1;
        (*tree)->x2 = rootline->x2;
        (*tree)->y2 = rootline->y2;

        (*tree)->f = NULL;
        (*tree)->n = NULL;

        if (frontlist != NULL)
        {
            if (BuildBSPTree(&((*tree)->n), frontlist, frontlist))
            {
                KillList(frontlist);
                KillList(backlist);
                return 1;
            }
        }

        if (backlist != NULL)
        {
            if (BuildBSPTree(&((*tree)->f), backlist, backlist))
            {
                KillList(frontlist);
                KillList(backlist);
                return 1;
            }
        }
    }
    else
        error_code = 1; // malloc() failed

    KillList(frontlist);
    KillList(backlist);
        
    return error_code;
}

/*
 * OutputBSPTree(): Outputs a BSP tree by recursively calling itself. When
 * reading a BSP file, the same procedure must be used to build the tree while
 * reading the file so that everything ends up in the proper order
 */

static void OutputBSPTreeNode(BSP *b)
{
    if (b == NULL)
        return;

    printf("%d,%d,%d:%d,%d-%d,%d\n", (int) b, (int) b->f, (int) b->n, b->x1, b->y1, b->x2, b->y2);
    OutputBSPTreeNode(b->f);
    OutputBSPTreeNode(b->n);
}

static void OutputBSPTree(BSP *b)
{
    OutputBSPTreeNode(b);
    printf("0,0,0:0,0-0,0\n");
}

/*
 * FreeBSPTree(): Frees all memory occupied by the BSP tree
 */

static void FreeBSPTree(BSP *b)
{
    if (b == NULL)
        return;

    FreeBSPTree(b->f);
    FreeBSPTree(b->n);

    free(b);
}


/*****************************************************************************
 DOS Code
*****************************************************************************/

/*
 * main()
 */

int main(int argc, char **argv)
{
    FILE    *fp;
    LINE    *l;

    if (argc <= 1)
    {
        puts("bsp: no input file, nothing to do");
        exit(0);
    }

    /*
     * Read in all of the lines
     */

    if ((fp = fopen(argv[1], "r")) == NULL)
    {
        fprintf(stderr, "bsp: unable to open %s\n", argv[1]);
        exit(0);
    }

    while (1)
    {
        int x1, y1, x2, y2;

        fscanf(fp, "%d,%d-%d,%d", &x1, &y1, &x2, &y2);
        if (x1 == 0 && y1 == 0 && x2 == 0 && y2 == 0)   // end of list!
            break;

        if ((l = malloc(sizeof(LINE))) == NULL)
        {
            fclose(fp);
            KillList(line_list);
            fprintf(stderr, "bsp: out of memory!\n");
            exit(1);
        }

        l->next = line_list;    // link into beginning of list
        l->x1 = x1;
        l->y1 = y1;
        l->x2 = x2;
        l->y2 = y2;

        line_list = l;
    }

    l = line_list;
    for (line_id = 0; l != NULL; line_id++) // assign a number to each element in the line list
    {
        l->id = line_id;
        l = l->next;
    }

    fclose(fp);

    /*
     * Build the tree
     */

    BuildBSPTree(&tree, line_list, line_list);

    /*
     * Print drawing order and then print out a BSP file
     */

    OutputBSPTree(tree);

    FreeBSPTree(tree);
    KillList(line_list);
    return 0;
}
