/*
 * poly.c
 *
 * Draws polygons
 */

#include <assert.h>
#include <stdlib.h>
#include "types.h"
#include "poly.h"


/*****************************************************************************
 Polygon Rasterization
*****************************************************************************/

/*
 * Macros for Vertex List Scanning (used by DrawPolygon())
 */

#define INDEX_FORWARD(index)    index++; if (index >= num_vertices) index = 0;
#define INDEX_BACKWARD(index)   index--; if (index < 0) index = num_vertices - 1;


/*
 * Edge Scanning
 *
 * Polygon edges are scanned and the end points are written to a span buffer
 * so as to render the horizontal strips later. This is where the real "magic"
 * of scan converting takes place.
 */

static void ScanEdge(INT32 x1, INT32 y1, INT32 x2, INT32 y2, INT32 left_edge, SPAN *span_list)
{
    UINT32  dx, dy;
    INT32   grad;   // fixed point

    /*
     * Sort the coordinates so that Y1 is less than Y2
     */

    if (y1 > y2)
    {
        dx = y1;    // use dx as temp variable since we aren't using it
        y1 = y2;
        y2 = dx;

        dx = x1;
        x1 = x2;
        x2 = dx;
    }

    /*
     * Calculate the X and Y deltas and the gradient
     */

    dx = abs(x2 - x1);
    dy = y2 - y1;
    if (!dy)
        return;

    grad = (dx << 16) / dy;
    if (x2 < x1)    // X will be going right-to-left
        grad = -grad;

    /*
     * Scan the edge!
     */

    x1 <<= 16;
    if (left_edge)
    {
        while (dy--)
        {
            (span_list)->xstart = x1;
            (span_list)++;
            x1 += grad;
        }
    }
    else
    {
        while (dy--)
        {
            (span_list)->xend = x1;
            (span_list)++;
            x1 += grad;
        }
    }  
}

/*
 * Span Drawing
 *
 * "Spans" are just a bunch of horizontal lines which form the polygon. They
 * are stored in a list which is created by the edge scanner and are drawn
 * here.
 */

static void DrawSpans(UINT32 y, UINT8 color, UINT32 num_spans, UINT32 xpitch, UINT8 *vidbuf, SPAN *span_list)
{
    UINT32  i, xstart, xend, tmp;
    UINT8   *v;
    
    /*
     * Find the first span which is not 1 pixel wide so that we can determine
     * whether or not all of the spans will be swapped
     */

    v = &vidbuf[y * xpitch];
    for (i = 0; i < num_spans; i++)
    {
        xstart = span_list[i].xstart >> 16;
        xend = span_list[i].xend >> 16;
        if (xstart != xend)
            break;
        v += xpitch;
    }

    /*
     * Convert the xstart and xend coordinates from 16.16 fixed point
     * format into regular integers
     */

    for (; i < num_spans; i++)
    {
        xstart = span_list[i].xstart >> 16;
        xend = span_list[i].xend >> 16;
        if (xstart > xend)
        {
            tmp = xstart;
            xstart = xend;
            xend = tmp;
        }

        /*
         * Draw the line quickly using memset(), and go on to the next line
         */
    
        memset(v + xstart, color, xend - xstart);   // draw the line
        v += xpitch;                                // next line
    }
}

/*
 * void Polygon_DrawSolid(INT32 xo, INT32 yo, UINT8 color, UINT32 xpitch,
 *                        UINT8 *vidbuf, VERTEX2D *v, UINT32 num_vertices);
 *
 * Draws a solid polygon. This function itself simply traverses the vertex
 * list in the proper order in order to scan edges properly and then calls a
 * span drawing function to actually render all of the spans.
 */

void Polygon_DrawSolid(INT32 xo, INT32 yo, UINT8 color, UINT32 xpitch, UINT8 *vidbuf, VERTEX2D *v, UINT32 num_vertices)
{
    SPAN    span_list[MAX_SPANS], *working_span_list_ptr;
    INT32   i, j, top_index, top_y, bottom_y;

    /*
     * Polygons with less than 3 vertices aren't really polygons
     */

    if (num_vertices < 3)
        return;

    /*
     * Find the top and bottom vertices
     */

    top_y = bottom_y = v[0].y;  // initialize to something valid
    top_index = 0;
    for (i = 1; i < num_vertices; i++)
    {
        if (v[i].y < top_y)
        {
            top_y = v[i].y;
            top_index = i;
        }
        else if (v[i].y > bottom_y)
            bottom_y = v[i].y;
    }

    /*
     * Now scan! First, we scan forward, then we scan backward.
     */

    working_span_list_ptr = span_list;
    i = top_index;
    do
    {
        j = i;
        INDEX_FORWARD(i);
        ScanEdge(v[j].x + xo, v[j].y,
                 v[i].x + xo, v[i].y,
                 0, working_span_list_ptr);
        working_span_list_ptr += abs(v[j].y - v[i].y);
    }
    while (v[i].y != bottom_y);

    working_span_list_ptr = span_list;
    i = top_index;
    do
    {
        j = i;
        INDEX_BACKWARD(i);
        ScanEdge(v[j].x + xo, v[j].y,
                 v[i].x + xo, v[i].y,
                 1, working_span_list_ptr);
        working_span_list_ptr += abs(v[j].y - v[i].y);
    }
    while (v[i].y != bottom_y);

    /*
     * Render the spans
     */

    DrawSpans(yo + top_y, color, bottom_y - top_y, xpitch, vidbuf, span_list);
}
